"""
模型调度器
负责多模型故障转移和负载均衡
"""

from typing import List, Dict, Any, Optional, Callable, Union
from llmakits.message import convert_to_json
from llmakits.load_model import load_models
from funcguard.printer import print_line, print_block


class ModelDispatcher:
    """
    模型调度器类，负责管理模型切换次数和执行任务
    """

    def __init__(
        self,
        models_config: Optional[Union[str, Dict[str, Any]]] = None,
        model_keys: Optional[Union[str, Dict[str, Any]]] = None,
    ):
        self.model_switch_count = 0
        self.exhausted_models = []

        if models_config and model_keys:
            self.model_groups, self.model_keys = load_models(models_config, model_keys)
        else:
            self.model_groups = {}
            self.model_keys = {}

    # 输出报告
    def report(self):
        print(f"Model switch count: {self.model_switch_count}")
        if self.exhausted_models:
            print(f"Exhausted models: {self.exhausted_models}")
        return

    # 移除Token已用尽的模型
    def _remove_exhausted_model(self, sdk_name: str, model_name: str):
        """
        从模型组中删除API密钥用尽的模型

        Args:
            sdk_name: 模型的SDK名称
            model_name: 模型的名称
        """
        # 从当前列表中删除匹配的模型
        for group_name, group_models in self.model_groups.items():
            new_group_models = []
            for model in group_models:
                if model['sdk_name'] != sdk_name or model['model_name'] != model_name:
                    new_group_models.append(model)
            self.model_groups[group_name] = new_group_models
        return

    # 执行任务 - 多模型调度器支持故障转移和重试
    def execute_task(
        self,
        message_info: Dict[str, Any],
        llm_models: List[Dict[str, Any]],
        format_json: bool = False,
        validate_func: Optional[Callable[[str], tuple[bool, Any]]] = None,
    ) -> tuple[Any, int]:
        """
        执行任务 - 多模型调度器支持故障转移和重试

        Args:
            message_info:
                { "system_prompt": "", "user_text": "", "include_img": False, "img_list": [] }
            llm_models: LLM模型列表，每个元素包含model、sdk_name等
            format_json: 是否格式化为JSON
            validate_func: 结果验证函数，可返回tuple[bool, Any]。
                        bool：True表示验证通过，False表示验证失败
                        Any：验证通过的值

        Returns:
            (返回消息, token总数)
        """
        models_num = len(llm_models)

        for idx, model_info in enumerate(llm_models):
            sdk_name = model_info.get('sdk_name', 'unknown_sdk')
            model_name = model_info.get('model_name', 'unknown_model')
            base_model_info = f"{idx+1}/{models_num} Model {sdk_name} : {model_name}"
            try:
                return_message, total_tokens = model_info["model"].send_message([], message_info)
                if format_json:
                    return_message = convert_to_json(return_message)

                # 如果有校验函数，校验不通过则继续下一个模型
                if validate_func is not None:
                    is_valid, validated_value = validate_func(return_message)

                    if is_valid:
                        if validated_value:
                            # 验证通过，直接返回验证通过的值
                            return validated_value, total_tokens
                    else:
                        content = "输出结果：条件校验失败, trying next model ..."
                        print_block(base_model_info, content)
                        self.model_switch_count += 1
                        continue

                return return_message, total_tokens

            except Exception as e:
                print_line()
                print(base_model_info)

                # 检查是否是API密钥用尽异常
                if str(e) == 'API_KEY_EXHAUSTED':
                    print(f"模型 {sdk_name} - {model_name} API密钥 已用完")
                    # 从模型组中删除该模型
                    self._remove_exhausted_model(sdk_name, model_name)
                    self.exhausted_models.append(f"{sdk_name}_{model_name}")

                if idx < models_num - 1:
                    print("model failed, trying next model ...")
                    print_line()
                    self.model_switch_count += 1
                    continue
                else:
                    raise e

        # 如果所有模型都失败
        raise Exception("All models failed.")

    # 执行任务 - 模型组调度器
    def execute_with_group(
        self,
        message_info: Dict[str, Any],
        group_name: str,
        format_json: bool = False,
        validate_func: Optional[Callable[[str], tuple[bool, Any]]] = None,
    ) -> tuple[Any, int]:
        """
        使用内部model_groups执行任务，避免重复实例化导致的状态丢失

        这是推荐的使用方式，因为：
        1. 模型实例在dispatcher中缓存，避免重复实例化
        2. 模型内部的API密钥切换状态会保持（同一个对象引用）
        3. 多次调用不会丢失已切换的密钥状态

        Args:
            message_info:
                { "system_prompt": "", "user_text": "", "include_img": False, "img_list": [] }
            group_name: 模型组名称
            format_json: 是否格式化为JSON
            validate_func: 结果验证函数

        Returns:
            (返回消息, token总数)
        """
        if not self.model_groups:
            raise Exception("没有可用的模型组，请先初始化dispatcher")

        # 获取模型列表（这些是缓存的模型实例，保持API密钥切换状态）
        llm_models = self.model_groups[group_name]
        if not llm_models:
            raise Exception(f"未找到模型组: {group_name}")

        # 使用现有的execute_task方法，传入group_name以便在API密钥用尽时删除模型
        return self.execute_task(message_info, llm_models, format_json, validate_func)



"""
Message处理模块
提供消息格式化、准备和验证功能
"""

from .builder import prepare_messages, rebuild_messages_single_image
from .formatter import convert_to_json, extract_field

__all__ = ['prepare_messages', 'rebuild_messages_single_image', 'convert_to_json', 'extract_field']



import re
from llmakits.dispatcher import ModelDispatcher


def check_allowed_tags(html_string: str, allowed_tags: set[str]):
    """
    检查HTML字符串中的标签是否都在允许的标签列表中。

    Args:
        html_string: 要检查的HTML字符串
        allowed_tags: 允许使用的标签集合

    Returns:
        set: 未被允许的标签名集合
    """
    # 查找所有HTML标签 (包括开始标签、结束标签和自闭合标签)
    # 这个正则表达式会匹配 <tag ...> 或 </tag> 或 <tag ... />
    found_tags = re.findall(r'<\s*/?([a-zA-Z]+)[^>]*>', html_string)

    # 将找到的标签名转换为小写集合以便比较
    found_tag_names = {tag.lower() for tag in found_tags}

    # 检查所有找到的标签是否都在允许的列表中
    unallowed_tags = found_tag_names - allowed_tags

    return unallowed_tags


def check_tag_closing(html_string: str):
    """
    检查HTML字符串中的标签是否正确闭合。

    Args:
        html_string: 要检查的HTML字符串

    Returns:
        set: 未正确闭合的标签名集合
    """
    # 自闭合标签列表
    self_closing_tags = {
        'br',
        'hr',
        'img',
        'input',
        'meta',
        'link',
        'area',
        'base',
        'col',
        'embed',
        'source',
        'track',
        'wbr',
    }

    # 匹配开始标签、结束标签和自闭合标签
    tag_pattern = r'<\s*(/?)([a-zA-Z]+)[^>]*>'
    tags = re.findall(tag_pattern, html_string)

    tag_stack = []
    unclosed_tags = set()

    for is_closing, tag_name in tags:
        tag_name = tag_name.lower()

        # 跳过自闭合标签
        if tag_name in self_closing_tags:
            continue

        if not is_closing:  # 开始标签
            tag_stack.append(tag_name)
        else:  # 结束标签
            if tag_stack and tag_stack[-1] == tag_name:
                tag_stack.pop()
            else:
                # 标签未正确闭合
                unclosed_tags.add(tag_name)

    # 栈中剩余的标签都是未闭合的
    unclosed_tags.update(tag_stack)
    return unclosed_tags


def validate_html(html_string: str, allowed_tags: set[str]):
    """
    校验HTML字符串是否仅包含指定的标签，并检查标签是否正确闭合。

    Args:
        html_string: 要校验的HTML字符串
        allowed_tags: 允许使用的标签集合

    Returns:
        tuple: (是否校验通过, 错误信息)
               如果校验通过，返回(True, "")
               如果校验失败，返回(False, 错误描述字符串)
    """

    # 构建错误信息
    error_messages = []

    # 检查标签是否被允许（仅当 allowed_tags 不为空时）
    if allowed_tags:
        unallowed_tags = check_allowed_tags(html_string, allowed_tags)
        if unallowed_tags:
            error_messages.append(f"发现未被允许的标签: {', '.join(sorted(unallowed_tags))}")

    # 检查标签闭合情况
    unclosed_tags = check_tag_closing(html_string)

    if unclosed_tags:
        error_messages.append(f"发现未正确闭合的标签: {', '.join(sorted(unclosed_tags))}")

    # 返回结果
    if error_messages:
        print(error_messages)
        return False, '; '.join(error_messages)

    return True, ""


def validate_html_fix(
    dispatcher: ModelDispatcher, html_string: str, allowed_tags: set[str], group_name: str, system_prompt: str
):
    """
    校验HTML字符串是否合规，并修复不允许的标签。
    """
    is_valid, error_messages = validate_html(html_string, allowed_tags)
    fixed_num = 0
    max_attempts = 5
    while not is_valid and fixed_num < max_attempts:
        fixed_num += 1
        message_info = {
            "system_prompt": system_prompt,
            "user_text": f"allowed_tags:{allowed_tags},html:{html_string},error_messages:{error_messages}",
        }
        return_message, _ = dispatcher.execute_with_group(message_info, group_name, format_json=True)
        html_string = return_message["html"]
        is_valid, error_messages = validate_html(html_string, allowed_tags)
    return html_string



"""
消息验证器
负责验证消息格式和内容的有效性
"""

from typing import Any, Dict, List


def validate_message_format(messages: List[Dict[str, Any]], provider_name: str) -> bool:
    """
    验证消息格式是否符合指定提供商的要求

    Args:
        messages: 消息列表
        provider_name: 提供商名称

    Returns:
        True如果格式有效，False否则
    """

    if len(messages) < 1:
        return False

    # 验证每条消息的基本结构
    for message in messages:
        if not isinstance(message, dict):
            return False

        if 'role' not in message or 'content' not in message:
            return False

        if message['role'] not in ['system', 'user', 'assistant']:
            return False

    # 根据提供商验证特定格式
    return _validate_provider_format(messages, provider_name)


def _validate_provider_format(messages: List[Dict[str, Any]], provider_name: str) -> bool:
    """验证特定提供商的消息格式"""

    if provider_name == "dashscope":
        return _validate_dashscope_format(messages)
    elif provider_name in ["zhipu", "openai", "modelscope"]:
        return _validate_openai_format(messages)
    elif provider_name == "ollama":
        return _validate_ollama_format(messages)
    else:
        return False


def _validate_dashscope_format(messages: List[Dict[str, Any]]) -> bool:
    """验证DashScope格式"""
    try:
        for message in messages:
            content = message['content']
            if isinstance(content, list):
                for item in content:
                    if not isinstance(item, dict):
                        return False
                    if 'text' not in item and 'image' not in item:
                        return False
            elif not isinstance(content, str):
                return False
        return True
    except Exception:
        return False


def _validate_openai_format(messages: List[Dict[str, Any]]) -> bool:
    """验证OpenAI格式"""
    try:
        for message in messages:
            content = message['content']
            if isinstance(content, list):
                for item in content:
                    if not isinstance(item, dict):
                        return False
                    if item.get('type') == 'image_url':
                        if 'image_url' not in item or not isinstance(item['image_url'], dict):
                            return False
                    elif item.get('type') == 'text':
                        if 'text' not in item:
                            return False
            elif not isinstance(content, str):
                return False
        return True
    except Exception:
        return False


def _validate_ollama_format(messages: List[Dict[str, Any]]) -> bool:
    """验证Ollama格式"""
    try:
        for message in messages:
            content = message['content']
            if not isinstance(content, str):
                return False
        return True
    except Exception:
        return False



from typing import Any, Dict, List, Union, Callable


def validate_dict(item_list: List[Dict[str, Any]], search_data: Dict[str, Any]) -> Dict[str, Any]:
    """
    验证并查找指定值是否存在于字典列表中

    :param item_list: (list[dict]): 要搜索的字典列表
    :param search_data: dict, 键名应与item_list中字典的键名一致
    :return: dict, 匹配到的项，包含与search_data相同的键
    """
    if not isinstance(item_list, list) or not all(isinstance(item, dict) for item in item_list):
        raise TypeError("item_list 必须是字典组成的列表")

    if not item_list:
        print("item_list 为空列表")
        return {}

    # 获取item_list中第一个字典的键名
    first_item_keys = set(item_list[0].keys())
    search_data_keys = set(search_data.keys())

    # 检查search_data的键是否与item_list中字典的键一致
    if not search_data_keys.issubset(first_item_keys):
        raise ValueError(f"search_data的键 {search_data_keys} 必须是item_list中字典键 {first_item_keys} 的子集")

    # 获取search_data的键和值
    search_keys = list(search_data.keys())
    search_values = [search_data[key] for key in search_keys]

    # 在item_list中查找匹配项
    for item in item_list:
        # 检查所有search_data的键值是否都匹配
        if all(item.get(key) == value for key, value in zip(search_keys, search_values)):
            # 返回完整的item
            return item

    # key没有匹配到时，尝试通过value进行匹配
    for item in item_list:
        # 检查是否存在与search_data值对应的键值对
        match_count = 0
        for search_value in search_values:
            for key, value in item.items():
                if value == search_value:
                    match_count += 1
                    break
        if match_count == len(search_values):
            # 返回完整的item
            return item

    # 未找到匹配项
    print(f"未找到匹配项: {search_data}")
    return {}


def validate_string(item_list: List[str], search_value: str):
    """
    验证并查找指定字符串是否存在于字符串列表中

    :param item_list: List[str], 要搜索的字符串列表
    :param search_value: str, 要查找的字符串值
    :return: str, 匹配到的字符串值
    """

    if search_value in item_list:
        return search_value

    # 未找到匹配项
    print(f"未找到匹配项: {search_value}")
    return None


def auto_validate(item_list: List[Any], search_data: Any) -> Union[Dict[str, Any], str, None]:
    """
    根据item_list的第一个元素类型自动选择验证程序

    Args:
        item_list: 要搜索的列表，可以是字典列表或字符串列表
        search_data: 要查找的数据，字典或字符串

    Returns:
        匹配到的值：字典验证返回Dict，字符串验证返回str，未找到返回None
    """
    if not item_list:
        print("item_list 为空列表")
        return None

    # 根据列表第一个元素的类型选择验证器
    first_item = item_list[0]

    if isinstance(first_item, dict):
        # 字典类型验证
        if not isinstance(search_data, dict):
            print(f"字典类型验证需要search_data为字典，实际类型: {type(search_data)}")
            return None
        return validate_dict(item_list, search_data)

    elif isinstance(first_item, str):
        # 字符串类型验证
        if not isinstance(search_data, str):
            print(f"字符串类型验证需要search_data为字符串，实际类型: {type(search_data)}")
            return None
        return validate_string(item_list, search_data)

    else:
        print(f"不支持的列表元素类型: {type(first_item)}")
        return None



"""
消息构建器
负责根据不同提供商的要求构建消息格式
"""

from typing import List, Optional, Dict, Any
from filekits.base_io.down_load import batch_download_encode_base64


def prepare_messages(
    provider_name: str,
    system_prompt: str,
    user_text: str,
    include_img: bool = False,
    img_list: Optional[List[str]] = None
) -> List[Dict[str, Any]]:
    """
    根据提供商名称准备消息格式
    
    Args:
        provider_name: 提供商名称 ('dashscope', 'zhipu', 'openai', 'modelscope', 'ollama')
        system_prompt: 系统提示词
        user_text: 用户文本
        include_img: 是否包含图片
        img_list: 图片URL列表
    
    Returns:
        格式化后的消息列表
    """
    if img_list is None:
        if include_img :
            raise ValueError("包含图片的消息，图片 img_list 不能为空。")
        img_list = []
    
    if include_img and len(img_list) < 1:
        raise ValueError("包含图片的消息，图片 img_list 不能为空。")
    
    # 根据提供商构建不同格式的消息
    system_content, user_content = _build_content_by_provider(
        provider_name, system_prompt, user_text, include_img, img_list
    )
    
    # 构建消息结构
    messages = _build_message_structure(provider_name, system_content, user_content, img_list)
    
    return messages


def rebuild_messages_single_image(
    provider_name: str,
    system_prompt: str,
    user_text: str,
    include_img: bool = False,
    img_list: Optional[List[str]] = None
) -> List[Dict[str, Any]]:
    """
    重新构造messages，只使用第一张图片
    
    Args:
        provider_name: 提供商名称
        system_prompt: 系统提示词
        user_text: 用户文本
        include_img: 是否包含图片
        img_list: 图片URL列表
    
    Returns:
        格式化后的消息列表（只包含第一张图片）
    """
    if img_list is None:
        img_list = []
    
    if len(img_list) == 1 and include_img :
        print("异常：图片数量 未超出限制")
        print(f"img_list : {img_list}")
        raise Exception("异常：图片数量 == 1，未超出限制")

    # 重新构造messages，只使用第一张图片
    return prepare_messages(
        provider_name, 
        system_prompt, 
        user_text, 
        include_img,
        [img_list[0]] if img_list else []
    )


def _build_content_by_provider(
    provider_name: str,
    system_prompt: str,
    user_text: str,
    include_img: bool,
    img_list: List[str]
) -> tuple:
    """根据提供商构建内容格式"""
    
    if not include_img:
        return system_prompt, user_text
    
    if provider_name == "dashscope":
        user_content = [{"image": img} for img in img_list]
        user_content.append({"text": user_text})
        system_content = [{"text": system_prompt}]
        
    elif provider_name in ["zhipu", "openai", "modelscope"]:
        user_content = [{"type": "image_url", "image_url": {"url": img}} for img in img_list]
        user_content.append({"type": "text", "text": user_text})
        system_content = [{"type": "text", "text": system_prompt}]
        
    elif provider_name == "ollama":
        system_content = system_prompt
        user_content = user_text
        img_list = batch_download_encode_base64(img_list)
        
    else:
        raise ValueError("Unsupported provider_name. Use 'dashscope', 'zhipu', 'openai', 'modelscope', or 'ollama'.")
    
    return system_content, user_content


def _build_message_structure(
    provider_name: str,
    system_content: Any,
    user_content: Any,
    img_list: List[str]
) -> List[Dict[str, Any]]:
    """构建消息结构"""
    
    system_message = {"role": "system", "content": system_content}
    user_message = {"role": "user", "content": user_content}
    
    # 为ollama添加图片base64编码
    if provider_name == "ollama" and len(img_list) > 0:
        user_message["images"] = img_list
    
    return [system_message, user_message]